"""Module with functions for computing distances between two sets of bounding boxes."""

import numba
import numpy as np
import numpy.typing as npt


@numba.njit(  # type: ignore[misc]
    numba.float32[:, ::1](numba.float32[:, ::1], numba.float32[:, ::1]),
    fastmath=True,
    parallel=True,
)
def iou_dist(
    bboxes_1: npt.NDArray[np.float32], bboxes_2: npt.NDArray[np.float32]
) -> npt.NDArray[np.float32]:
    """
    Compute the IoU distance between two batches of bounding boxes.

    The IoU distance is computed as 1 minus the IoU similarity.

    Args:
        bboxes_1: An `[N, 4]` array of bounding boxes in the xywh format.
        bboxes_2: An `[M, 4]` array of bounding boxes in the xywh format.

    Returns:
        An `[N, M]` array of pairwise IoU distances.

    """
    ious = np.zeros((bboxes_1.shape[0], bboxes_2.shape[0]), dtype=np.float32)

    for i in numba.prange(bboxes_1.shape[0]):
        for j in numba.prange(bboxes_2.shape[0]):
            xx1 = max(bboxes_1[i, 0], bboxes_2[j, 0])
            xx2 = min(bboxes_1[i, 0] + bboxes_1[i, 2], bboxes_2[j, 0] + bboxes_2[j, 2])
            yy1 = max(bboxes_1[i, 1], bboxes_2[j, 1])
            yy2 = min(bboxes_1[i, 1] + bboxes_1[i, 3], bboxes_2[j, 1] + bboxes_2[j, 3])

            zero = numba.float32(0)
            intersection = max(zero, xx2 - xx1) * max(zero, yy2 - yy1)
            if intersection > 0:
                union = (
                    bboxes_1[i, 2] * bboxes_1[i, 3]
                    + bboxes_2[j, 2] * bboxes_2[j, 3]
                    - intersection
                )
                ious[i, j] = intersection / union

    return 1 - ious



"""Module containing a util function for computing the precision-recall curve."""

import numpy as np
import numpy.typing as npt


def prec_recall_curve(
    hyp_matched: npt.NDArray[np.bool_],
    hyp_conf: npt.NDArray[np.float32],
    n_gts: int,
) -> tuple[npt.NDArray[np.float64] | None, npt.NDArray[np.float64] | None]:
    """
    Calculate the precision-recall curve.

    This calculates the precision-recall curve for matched hypotheses and ground truths.
    There are two possible edge cases:
    * when there are no ground truths: in this case `None` is returned
    * when there are no hypotheses (predictions): in this case the PR curve consists
      of a single `(0,0)` point.

    Args:
        hyp_matched: A `[N,]` array indicating whether a hypotheses is matched or not.
        hyp_conf: A `[N,]` array of hypotheses' confidence scores.
        n_gts: Number of ground truths

    Returns:
        If `n_gts == 0`, the function returns `None`. Otherwise two `L` length arrays
        are returned: `precision` and `recall`. The contain all points on the precision
        recall curve.

        If there are no hypotheses (`N==0`), by convention the precision-recall curve
        will contain only the `(0,0)` point.

    """
    if n_gts == 0:
        return None, None

    if len(hyp_matched) == 0:
        return (np.array([0.0]), np.array([0]))

    hyp_matched_cum_sum = np.cumsum(
        hyp_matched[np.argsort(-hyp_conf, kind="mergesort")]
    )
    precision = hyp_matched_cum_sum / np.arange(1, len(hyp_matched) + 1)
    recall = hyp_matched_cum_sum / n_gts

    return precision, recall



"""Various utils for computing MOT metrics."""

import numpy as np
from scipy import sparse
from scipy.optimize import linear_sum_assignment

from evaldet import Tracks
from evaldet.dist import iou_dist


def create_coo_array(
    vals_dict: dict[tuple[int, int], int], shape: tuple[int, int]
) -> sparse.coo_array:
    """
    Create a sparse COO array.

    Args:
        vals_dict: A dictionary with values. The key should be a tuple of
            ``(row_ind, col_ind)``, and the value should be the entry for the cell
            at that index.
        shape: Shape of the new array: ``(n_rows, n_cols)``

    """
    row_inds = np.array(tuple(x[0] for x in vals_dict))
    col_inds = np.array(tuple(x[1] for x in vals_dict))
    vals = np.array(tuple(vals_dict.values()))

    return sparse.coo_array((vals, (row_inds, col_inds)), shape=shape)


def preprocess_mot_1720(
    gt: Tracks, hyp: Tracks, mot_20: bool = True
) -> tuple[Tracks, Tracks]:
    """A utility for pre-processing MOT 17 and 20 files. Used in testing."""
    distractor_cls_ids = [2, 7, 8, 12]
    if mot_20:
        distractor_cls_ids.append(6)

    pedestrian_class = [1]

    hyp_filter = np.ones_like(hyp.ids, dtype=np.bool_)
    gt_filter = np.ones_like(gt.ids, dtype=np.bool_)

    all_frames = sorted(set(gt.frames).intersection(hyp.frames))
    for frame in all_frames:
        dist_matrix = iou_dist(gt[frame].bboxes, hyp[frame].bboxes)
        dist_matrix[dist_matrix > 0.5] = 1  # noqa: PLR2004
        matches_row, matches_col = linear_sum_assignment(dist_matrix)

        matches_filter = dist_matrix[matches_row, matches_col] < 0.5  # noqa: PLR2004
        matches_row = matches_row[matches_filter]
        matches_col = matches_col[matches_filter]

        classes = gt[frame].classes

        is_distractor = np.isin(
            classes[matches_row], distractor_cls_ids, assume_unique=True
        )
        to_remove_hyp = np.isin(
            np.arange(len(hyp[frame].ids)),
            matches_col[is_distractor],
            assume_unique=True,
        )

        start, end = hyp._frame_ind_dict[frame]  # noqa: SLF001
        hyp_filter[start:end] = ~to_remove_hyp

    gt_filter = gt_filter & (gt.confs >= 0.1)  # noqa: PLR2004
    gt_filter = gt_filter & np.isin(gt.classes, pedestrian_class)

    gt = gt.filter(gt_filter)
    hyp = hyp.filter(hyp_filter)

    return gt, hyp



"""
Module defining a general class for holding object detections,
as well as reading from and converting to various format.
"""

import datetime as dt
import json
from collections.abc import Sequence
from pathlib import Path
from typing import Any

import numpy as np
import numpy.typing as npt
import pyarrow as pa
import pyarrow.parquet as pq

_SEP = b"\x00"


class Detections:
    """
    A class representing object detections (with bounding boxes).

    It can read and save to the following file formats

    - YOLO format (as described [here](https://opencv.github.io/cvat/docs/manual/advanced/formats/format-yolo/))
    - COCO format (as described [here](https://cocodataset.org/#format-data))
    - PASCAL VOC format (as described [here](https://opencv.github.io/cvat/docs/manual/advanced/formats/format-voc/))
    - our custom efficient parquet format (see `from_parquet` and `to_parquet`)

    Internally, all the attributes are saved as a single numpy array, and sorted by
    image ids. This enables efficient storage while maintaining easy access to
    detections from individual frames.
    """

    _image_ids: npt.NDArray[np.int32]
    _bboxes: npt.NDArray[np.float32]
    _classes: npt.NDArray[np.int32]
    _confs: npt.NDArray[np.float32] | None
    _image_ind_dict: dict[int, tuple[int, int]]
    _iscrowd: npt.NDArray[np.bool_] | None

    _class_names: tuple[str, ...]
    _image_names: tuple[str, ...]

    def __init__(
        self,
        image_ids: npt.NDArray[np.int32] | Sequence[int],
        bboxes: Sequence[npt.NDArray[np.float32]] | npt.NDArray[np.float32],
        classes: npt.NDArray[np.int32] | Sequence[int],
        class_names: Sequence[str],
        image_names: Sequence[str],
        confs: npt.NDArray[np.float32] | Sequence[float] | None = None,
        iscrowd: npt.NDArray[np.bool_] | Sequence[bool] | None = None,
    ) -> None:
        """
        Create a `Detections` instance.

        Args:
            image_ids: A sequence or array of image ids. It is assumed to be 0-indexed.
            bboxes: A sequence or array of detection bounding boxes, which should be
                in the format `xywh`, using a top-left-origin coordinate system.
            classes: A sequence or array of classes (labels), assumed to be 0-indexed.
            confs: An optional sequence or array of confidences (scores).
            class_names: A sequence (list, tuple) of class names. The `i`-th element in
                the sequence (zero indexed) is the class name for detections with class
                label `i`. The length should be larger than the max class label in
                `classes`.
            image_names: A sequence (list, tuple) of image names - this is used for
                matching ground truth and prediction detection for metrics. The `i`-th
                element in the sequence (zero indexed) is the image name for all images
                with the label `i`. The length should be larger than the max image label
            iscrowd: A sequence of booleans that determine if the detection is a crowd.
                This is only meant to be used for COCO dataset ground truths, as it is
                required to properly compute matching and COCO metrics. If not provided,
                all detections are assumed not to be a crowd.

        """
        self._check_lengths(image_ids, bboxes, classes, confs, iscrowd)

        if len(image_ids) == 0:
            self._image_ids = np.zeros((0,), dtype=np.int32)
            self._bboxes = np.zeros((0, 4), dtype=np.float32)
            self._classes = np.zeros((0,), dtype=np.int32)
            self._confs = np.zeros((0,), dtype=np.float32)
            self._iscrowd = np.zeros((0,), dtype=np.bool_)

        else:
            image_ids = np.array(image_ids)

            sort_inds = np.argsort(image_ids, kind="stable")

            self._image_ids = np.array(image_ids[sort_inds], copy=True, dtype=np.int32)

            self._bboxes = np.array(
                np.array(bboxes)[sort_inds], dtype=np.float32, copy=True
            )

            self._classes = np.array(
                np.array(classes)[sort_inds], dtype=np.int32, copy=True
            )

            if confs is None:
                self._confs = None
            else:
                self._confs = np.array(
                    np.array(confs)[sort_inds], dtype=np.float32, copy=True
                )

            if iscrowd is None:
                self._iscrowd = None
            else:
                self._iscrowd = np.array(
                    np.array(iscrowd)[sort_inds], dtype=np.bool_, copy=True
                )

        # Check that class_names cover all classes
        num_classes = self._classes.max(initial=-1) + 1
        if len(class_names) < num_classes:
            msg = (
                f"The number of class names ({len(class_names)}) is less than"
                f" the number of classes in the data ({num_classes})"
            )
            raise ValueError(msg)

        self._class_names = tuple(class_names)

        # Check that image_names cover all images
        num_images = self._image_ids.max(initial=-1) + 1
        if len(image_names) < num_images:
            msg = (
                f"The number of image names ({len(image_names)}) is less than the"
                f"number of images in the data ({num_images})"
            )
            raise ValueError(msg)

        self._image_names = tuple(image_names)

        self._create_image_ind_dict()

    def _check_lengths(
        self,
        image_ids: npt.NDArray[np.int32] | Sequence[int],
        bboxes: Sequence[npt.NDArray[np.float32]] | npt.NDArray[np.float32],
        classes: npt.NDArray[np.int32] | Sequence[int],
        confs: npt.NDArray[np.float32] | Sequence[float] | None = None,
        iscrowd: npt.NDArray[np.bool_] | Sequence[bool] | None = None,
    ) -> None:
        if len(image_ids) != len(bboxes):
            raise ValueError(
                "`image_ids` and `bboxes` should contain the same number of items."
            )

        if len(image_ids) != len(classes):
            raise ValueError(
                "`image_ids` and `classes` should contain the same number of items."
            )

        if confs is not None and len(confs) != len(image_ids):
            msg = (
                "If `confs` is given, it should contain the same number of items"
                " as `image_ids`."
            )
            raise ValueError(msg)

        if iscrowd is not None and len(iscrowd) != len(image_ids):
            msg = (
                "If `iscrowd` is given, it should contain the same number of items"
                " as `image_ids`."
            )
            raise ValueError(msg)

        if len(bboxes) > 0 and bboxes[0].shape != (4,):
            msg = (
                "Each row of `bboxes` should be an 4-item array, but got"
                f" shape {bboxes[0].shape}"
            )
            raise ValueError(msg)

    def _create_image_ind_dict(self) -> None:
        if len(self._image_ids) == 0:
            self._image_ind_dict = {}
            return

        u_image_ids, start_inds = np.unique(self._image_ids, return_index=True)
        image_start_inds = start_inds.tolist()
        image_end_inds = [*start_inds[1:].tolist(), len(self._image_ids)]

        image_start_end_inds = zip(image_start_inds, image_end_inds, strict=True)
        self._image_ind_dict = dict(
            zip(u_image_ids.tolist(), image_start_end_inds, strict=True)
        )

    @staticmethod
    def strs_to_bytes(strs: tuple[str, ...]) -> bytes:
        """
        Convert a tuple of strings into a single bytes object joined by the separator.
        """
        return _SEP.join(x.encode("utf8") for x in strs)

    @staticmethod
    def bytes_to_strs(strs: bytes) -> tuple[str, ...]:
        """
        Convert a bytes object, separated by the separator, back into a tuple of
        strings.
        """
        return tuple(x.decode("utf8") for x in strs.split(_SEP))

    @classmethod
    def from_coco(cls, file_path: str | Path) -> "Detections":
        """
        Read the detections from a json file in the COCO format.

        The file should have this structure:

        ```
        {
            "categories": [
                {
                    "id": 1,
                    "name": "cat"
                },
                ...
            ],
            "images": [
                {
                    "id": 1,
                    "file_name": "cat1.jpg"
                },
                ...
            ],
            "annotations": [
                {
                    "id": 1,
                    "image_id": 1,
                    "category_id": 1,
                    "bbox": [260, 177, 231, 199],
                    "score": 0.67,
                    "iscrowd": false
                },
            ]
        }
        ```

        Other information (like `info` and `licenses` sections, or information about
        `width` or `height` for `images`) can be present too, but it will be ignored.

        The `score` attribute of `annotations` is optional, but if present should be
        present on all annotations.

        Because `Detections` requires that images and classes are zero indexed, while
        in COCO there is no restriction on numbering of `id`s, the `id`s for `image`
        and `category` are not perserved - but the names and relationships are.

        Args:
            file_path: Path where the detections file is located

        """
        with open(file_path) as f:
            coco_dict: dict[str, Any] = json.load(f)

        images: list[dict[str, int | str]] = coco_dict["images"]
        images = sorted(images, key=lambda x: x["id"])

        categories: list[dict[str, int | str]] = coco_dict["categories"]
        annotations: list[dict[str, Any]] = coco_dict["annotations"]

        img_ind_dict = {im["id"]: i for i, im in enumerate(images)}
        cat_ind_dict = {cat["id"]: i for i, cat in enumerate(categories)}

        image_ids = np.zeros((len(annotations),), dtype=np.int32)
        classes = np.zeros((len(annotations),), dtype=np.int32)
        bboxes = np.zeros((len(annotations), 4), dtype=np.float32)
        iscrowd = np.zeros((len(annotations),), dtype=np.bool_)

        if len(annotations) == 0 or "score" in annotations[0]:
            confs = np.zeros((len(annotations),), dtype=np.float32)
        else:
            confs = None

        for i, ann in enumerate(annotations):
            image_ids[i] = img_ind_dict[ann["image_id"]]
            classes[i] = cat_ind_dict[ann["category_id"]]
            bboxes[i] = ann["bbox"]

            iscrowd[i] = bool(ann.get("iscrowd", 0))

            if confs is not None:
                confs[i] = ann["score"]

        image_names: tuple[str, ...] = tuple(im["file_name"] for im in images)  # type: ignore[misc]
        class_names: tuple[str, ...] = tuple(cat["name"] for cat in categories)  # type: ignore[misc]

        return cls(
            image_ids=image_ids,
            bboxes=bboxes,
            confs=confs,
            classes=classes,
            class_names=class_names,
            image_names=image_names,
            iscrowd=iscrowd,
        )

    # @classmethod
    # def from_yolo(cls) -> "Detections":
    #     pass

    # @classmethod
    # def from_pascal_voc(cls) -> "Detections":
    #     pass

    @classmethod
    def from_parquet(cls, file_path: str | Path) -> "Detections":
        """
        Read the detections from a parquet file.

        The file should have the following columns:

        ```
        image_id, xmin, ymin, width, height, class
        ```

        Additionally, it can have a `conf` column. In metadata, it can provide the
        `class_names` and `image_names` attributes (concatenated names separated by a
        special character, as done in `to_parquet`).

        Args:
            file_path: Path where the detections file is located

        """
        table = pq.read_table(file_path)

        image_ids = table["image_id"].to_numpy()
        x = table["xmin"].to_numpy()
        y = table["ymin"].to_numpy()
        w = table["width"].to_numpy()
        h = table["height"].to_numpy()
        classes = table["class"].to_numpy()

        bboxes = np.stack([x, y, w, h]).T.copy()

        confs = table["conf"].to_numpy() if "conf" in table.column_names else None

        metadata = table.schema.metadata
        class_names = cls.bytes_to_strs(metadata[b"class_names"])
        image_names = cls.bytes_to_strs(metadata[b"image_names"])

        return cls(
            image_ids=image_ids,
            bboxes=bboxes,
            classes=classes,
            confs=confs,
            class_names=class_names,
            image_names=image_names,
        )

    def filter(self, mask: npt.NDArray[np.bool_]) -> "Detections":
        """
        Filter the detections using a boolean mask.

        This method will filter all attributes according to the mask provided.

        Args:
            mask: A boolean array, should be the same length as ids and other
                attributes.

        Return:
            The filtered detections

        """
        if mask.shape != self._image_ids.shape:
            msg = (
                "Shape of the filter should equal the shape of `image_ids`, got"
                f" {mask.shape}"
            )
            raise ValueError(msg)

        filtered_confs = self.confs if self.confs is None else self.confs[mask]
        filtered_iscrowd = self.iscrowd if self.iscrowd is None else self.iscrowd[mask]

        return Detections(
            image_ids=self.image_ids[mask],
            bboxes=self.bboxes[mask],
            classes=self.classes[mask],
            confs=filtered_confs,
            iscrowd=filtered_iscrowd,
            class_names=self.class_names,
            image_names=self.image_names,
        )

    def to_coco(self, file_path: str | Path) -> None:
        """
        Export detections to COCO format.

        Args:
            file_path: Relative or absolute file name to save to.

        """
        if self.image_names is None:
            num_images = self.image_ids.max(initial=-1) + 1
            images = [{"id": i, "file_name": f"{i}"} for i in range(num_images)]
        else:
            images = [
                {"id": i, "file_name": name} for i, name in enumerate(self.image_names)
            ]

        if self.class_names is None:
            num_categories = self.classes.max(initial=-1) + 1
            categories = [{"id": int(i), "name": str(i)} for i in range(num_categories)]
        else:
            categories = [
                {"id": i, "name": name} for i, name in enumerate(self.class_names)
            ]

        annotations: list[dict[str, Any]] = []
        for i in range(len(self.bboxes)):
            ann: dict[str, Any] = {}
            ann["image_id"] = int(self.image_ids[i])
            ann["category_id"] = int(self.classes[i])
            ann["bbox"] = [round(x, 3) for x in self.bboxes[i].tolist()]

            if self.confs is not None:
                ann["score"] = round(float(self.confs[i]), 3)

            annotations.append(ann)

        coco_dict = {
            "licences": [],
            "info": {
                "year": dt.datetime.now().year,
                "date_created": dt.datetime.now().isoformat(timespec="seconds"),
            },
            "images": images,
            "categories": categories,
            "annotations": annotations,
        }

        with open(file_path, "w+") as f:
            json.dump(coco_dict, f)

    # def to_pascal_voc(self) -> None:
    #     pass

    # def to_yolo(self) -> None:
    #     pass

    def to_parquet(self, file_path: str | Path) -> None:
        """
        Export detections to parquet format.

        Args:
            file_path: Relative or absolute file name to save to.

        """
        columns: list[pa.Array] = [
            pa.array(self._image_ids),
            pa.array(self._bboxes[:, 0]),
            pa.array(self._bboxes[:, 1]),
            pa.array(self._bboxes[:, 2]),
            pa.array(self._bboxes[:, 3]),
            pa.array(self._classes),
        ]
        col_names: list[str] = ["image_id", "xmin", "ymin", "width", "height", "class"]

        if self._confs is not None:
            columns.append(pa.array(self._confs))
            col_names.append("conf")

        table = pa.Table.from_arrays(columns, col_names)

        metadata: dict[bytes, bytes] = {}
        if self._class_names is not None:
            metadata[b"class_names"] = self.strs_to_bytes(self._class_names)

        if self._image_names is not None:
            metadata[b"image_names"] = self.strs_to_bytes(self._image_names)

        table = table.cast(table.schema.with_metadata(metadata))
        pq.write_table(table, file_path)

    @property
    def num_images(self) -> int:
        """Number of images with detections."""
        return len(self._image_ind_dict)

    @property
    def num_dets(self) -> int:
        """Number of detections."""
        return len(self._bboxes)

    @property
    def num_classes(self) -> int:
        """Number of classes with detections."""
        return len(np.unique(self._classes))

    @property
    def image_ids(self) -> npt.NDArray[np.int32]:
        """The array of image IDs associated with each bounding box."""
        return self._image_ids

    @property
    def bboxes(self) -> npt.NDArray[np.float32]:
        """The array of bounding boxes, each defined by [x_min, y_min, x_max, y_max]."""
        return self._bboxes

    @property
    def classes(self) -> npt.NDArray[np.int32]:
        """The array of integer class labels for each bounding box."""
        return self._classes

    @property
    def confs(self) -> npt.NDArray[np.float32] | None:
        """The array of confidence scores if available, otherwise None."""
        return self._confs

    @property
    def iscrowd(self) -> npt.NDArray[np.bool_] | None:
        """Optional COCO iscrowd flag per GT annotation (len == num_dets)."""
        return self._iscrowd

    @property
    def image_ind_dict(self) -> dict[int, tuple[int, int]]:
        """
        A dictionary mapping image IDs to start and end indices for all
        bounding boxes belonging to the image.
        """
        return self._image_ind_dict

    @property
    def class_names(self) -> tuple[str, ...]:
        """A tuple of class names corresponding to the integer class labels."""
        return self._class_names

    @property
    def image_names(self) -> tuple[str, ...]:
        """A tuple of image names corresponding to the image IDs."""
        return self._image_names



"""ID family of MOT metrics."""

import collections as co
import typing as t

import numpy as np
import numpy.typing as npt
from scipy.optimize import linear_sum_assignment

from evaldet.tracks import Tracks

from .utils import create_coo_array


class IDResults(t.TypedDict):
    """
    A typed dictionary for storing the results of the ID metric evaluation.
    """

    IDTP: float
    IDFP: float
    IDFN: float
    IDP: float
    IDR: float
    IDF1: float


def calculate_id_metrics(
    ground_truth: Tracks,
    hypotheses: Tracks,
    ious: dict[int, npt.NDArray[np.float32]],
    dist_threshold: float = 0.5,
) -> IDResults:
    """
    Calculate ID (identity) metrics.

    Args:
        ground_truth: Ground truth tracks.
        hypotheses: Hypotheses tracks.
        ious: A dictionary where keys are frame numbers (indices), and values
            are numpy matrices of IOU distances between detection in ground truth and
            hypotheses for that frame. IOUs must be present for all frames that are
            present in both ground truth and hypotheses.
        dist_threshold: The distance threshold for the computation of the metrics - used
            to determine whether to match two objects.

    Returns:
        A dictionary containing ID metrics:

            - IDP (ID Precision)
            - IDR (ID Recall)
            - IDF1 (ID F1)
            - IDFP (ID false positives)
            - IDFN (ID false negatives)
            - IDTP (ID true positives)

    """
    gts = tuple(ground_truth.ids_count.keys())
    gts_id_ind_dict = {_id: ind for ind, _id in enumerate(gts)}
    hyps = tuple(hypotheses.ids_count.keys())
    hyps_id_ind_dict = {_id: ind for ind, _id in enumerate(hyps)}

    gts_counts = np.array(tuple(ground_truth.ids_count.values()), dtype=np.int32)
    hyps_counts = np.array(tuple(hypotheses.ids_count.values()), dtype=np.int32)

    matches: dict[tuple[int, int], int] = co.defaultdict(int)
    for frame in sorted(set(ground_truth.frames).intersection(hypotheses.frames)):
        dist_matrix = ious[frame]
        gt_frame_inds = [gts_id_ind_dict[_id] for _id in ground_truth[frame].ids]
        htp_frame_inds = [hyps_id_ind_dict[_id] for _id in hypotheses[frame].ids]

        for gt_ind, hyp_ind in np.argwhere(dist_matrix < dist_threshold):
            matches[(gt_frame_inds[gt_ind], htp_frame_inds[hyp_ind])] += 1

    # Calculate matching as a LAP, get FN, FP and TP from matched entries
    matches_matrix = create_coo_array(matches, (len(gts), len(hyps)))
    matches_array = matches_matrix.toarray()
    row_m_inds, col_m_inds = linear_sum_assignment(matches_array, maximize=True)

    # Calculate the final results
    TPs = matches_array[row_m_inds, col_m_inds].sum()
    FNs = gts_counts.sum() - TPs
    FPs = hyps_counts.sum() - TPs

    idp = TPs / (TPs + FPs)
    idr = TPs / (TPs + FNs)
    idf1 = 2 * TPs / (2 * TPs + FPs + FNs)

    return {
        "IDTP": TPs,
        "IDFP": FPs,
        "IDFN": FNs,
        "IDP": idp,
        "IDR": idr,
        "IDF1": idf1,
    }



"""COCO object detection metrics."""

from collections.abc import Mapping
from enum import Enum
from typing import Any, TypedDict, cast

import numba
import numpy as np
import numpy.typing as npt
from numba import types

from evaldet import Detections
from evaldet.utils.prec_recall import prec_recall_curve

from . import utils


def _nonemean(x: npt.NDArray[np.float32]) -> float | None:
    if x.size == 0 or np.all(np.isnan(x)):
        return None

    return float(np.nanmean(x))


_BASE_50_95_THRESHOLDS = (
    0.5,
    0.55,
    0.6,
    0.65,
    0.7,
    0.75,
    0.8,
    0.85,
    0.9,
    0.95,
)


class COCOResult(TypedDict):
    """
    A typed dictionary for storing the results of the per-class COCO metric evaluation.
    """

    ap: float | None
    recall: float | None
    precision: float | None
    n_gts: int


class COCOResults(TypedDict):
    """
    A typed dictionary for storing the results of the COCO metric evaluation
    (across all classes).
    """

    mean_ap: float | None
    class_results: dict[str, COCOResult]


class COCOSummaryResults(TypedDict):
    """
    A typed dictionary for storing the COCO summary results.
    """

    mean_ap: float | None
    ap_50: float | None
    ap_75: float | None

    mean_ap_per_class: dict[str, float | None]
    ap_50_per_class: dict[str, float | None]
    ap_75_per_class: dict[str, float | None]

    mean_ap_sizes: dict[str, float | None]
    mean_ap_sizes_per_class: dict[str, dict[str, float | None]]


class APInterpolation(str, Enum):
    """
    Enum for specifying the method of Average Precision (AP) interpolation.
    """

    coco = "coco"
    pascal = "pascal"


@numba.njit(  # type: ignore[misc]
    types.void(
        types.float32[:, ::1],  # ious
        types.float32[::1],  # preds_area
        types.bool_[::1],  # gts_crowd_sorted
        types.float32[::1],  # gts_area_sorted
    ),
)
def _adjust_ious_for_crowd(
    ious: npt.NDArray[np.float32],
    preds_area: npt.NDArray[np.float32],
    gts_crowd_sorted: npt.NDArray[np.bool_],
    gts_area_sorted: npt.NDArray[np.float32],
) -> None:
    """
    Adjust IoU values for crowd ground truths following the COCO evaluation rule.

    In COCO, for crowd ground truths the similarity is defined as:

        iou(gt, dt, iscrowd=1) = area(intersect(gt, dt)) / area(dt)

    The input `ious` contains *standard* IoU:

        IoU = area(intersect) / area(union)

    For crowd ground truths, this function:
    * recovers the intersection area from the standard IoU, prediction area, and ground
      truth area, and then
    * replaces `ious[:, j]` with `intersection / preds_area` for each crowd ground truth
      column `j`.

    This function mutates `ious` in place.

    Args:
        ious: A `[N, M]` array of standard IoU similarities between predictions and
            ground truths (will be modified in place).
        preds_area: A `[N, ]` array with areas of prediction bounding boxes.
        gts_crowd_sorted: A `[M, ]` boolean array indicating which ground truths are
            crowd ground truths (after any sorting applied by the caller).
        gts_area_sorted: A `[M, ]` array with areas of ground truth bounding boxes
            (aligned with `gts_crowd_sorted`).

    Returns:
        None. The `ious` array is modified in place.

    """
    if np.any(gts_crowd_sorted):
        eps = np.float32(1e-10)
        for j in range(gts_crowd_sorted.shape[0]):
            if not gts_crowd_sorted[j]:
                continue
            for i in range(ious.shape[0]):
                iou = ious[i, j]
                if iou <= 0.0:
                    ious[i, j] = 0.0
                    continue
                inter = (
                    iou
                    * (preds_area[i] + gts_area_sorted[j])
                    / (np.float32(1.0) + iou + eps)
                )
                ious[i, j] = inter / (preds_area[i] + eps)


@numba.njit(  # type: ignore[misc]
    types.Tuple((types.int32[::1], types.bool_[::1], types.bool_[::1], types.int32))(
        types.float32[:, ::1],
        types.float32[:, ::1],
        types.bool_[::1],
        types.float32[:, ::1],
        types.float32[::1],
        types.Tuple((types.float32, types.float32)),
        types.float32,
    ),
)
def _evaluate_image(
    preds_bbox: npt.NDArray[np.float32],
    gts_bbox: npt.NDArray[np.float32],
    gts_crowd: npt.NDArray[np.bool_],
    ious: npt.NDArray[np.float32],
    preds_conf: npt.NDArray[np.float32],
    area_range: tuple[float, float],
    iou_threshold: float,
) -> tuple[npt.NDArray[np.int32], npt.NDArray[np.bool_], npt.NDArray[np.bool_], int]:
    """
    Evaluate a single image by matching predictions and ground truths for it.

    The matching process is the following:
    * First, ground truths are split into normal and ignored ones (those with area
      outside `area_range` or that are marked as a crowd)
    * In descending order according to prediction confidence, each prediction is matched
      with a ground truth that has not been matched to a previous prediction, and that
      has the highest IoU with it - the IoU needs to be above `iou_threshold`. If no
      match can be made with a normal ground truth, match is attempted with an ignored
      ground truth.
    * Finally, the predictions that were matched to an ignored ground truth, or those
      that were unmatched and had area outside of `area_range`, are marked as ignored.
      A ignored ground truth can only match with one detection - except for crowd ground
      truths, that can match any number of detections.

    Args:
        preds_bbox: A `[N, 4]` array of prediction bounding boxes in xywh format
        gts_bbox: A `[M, 4]` array of ground truth bounding boxes in xywh format
        gts_crowd: A `[M,]` array determining if a ground truth is a crowd or not.
        ious: A `[N, M]` array of IoU similarities between predictions and ground truths
        preds_conf: A `[N, ]` array of confidence scores of predictions
        area_range: A tuple `(lower_limit, upper_limit)` of limits for bounding box
            areas. Ground truths with area outside of this range will be ignored, as
            will the predictions matched to them and unmatched predictions with area
            outside of this range
        iou_threshold: The lower threshold for an IOU between a prediction and a
            bounding box that is required for establishing a match

    Returns:
        A tuple with 4 elements:
            - an integer array of shape `[N, ]`, indicating the index of the ground
              truth that the prediction was matched to. Unmatched predictions will have
              the matched index of `-1`.
            - a boolean array of shape `[N, ]`, indicating whether the prediction was
              ignored or not
            - a boolean array of shape `[M, ]`, indicating whether the ground truth was
              ignored or not
            - an integer denoting the number of non-ignored ground truths in the image

    """
    matched = np.full(preds_conf.shape, -1, dtype=np.int32)

    # Compute area
    gts_area = gts_bbox[:, 2] * gts_bbox[:, 3]
    preds_area = preds_bbox[:, 2] * preds_bbox[:, 3]

    ignore_preds_area = (preds_area < area_range[0]) | (preds_area > area_range[1])

    # Sort gts by ignore (COCO: crowd gts are treated as ignored)
    gts_ignore = (gts_area < area_range[0]) | (gts_area > area_range[1]) | gts_crowd
    sort_gt = np.argsort(gts_ignore, kind="mergesort")
    n_gts = int((~gts_ignore).sum())

    if preds_conf.size == 0 or gts_bbox.size == 0:
        return (matched, ignore_preds_area, gts_ignore, n_gts)

    # Sort preds by conf
    sort_preds = np.argsort(-preds_conf, kind="mergesort")

    ious = ious[:, sort_gt]
    gts_crowd_sorted = gts_crowd[sort_gt]
    gts_area_sorted = gts_area[sort_gt]

    # COCO crowd IoU: iou(gt, dt, iscrowd=1) = area(intersect(gt,dt)) / area(dt)
    # Here `ious` is standard IoU, so recover intersection and convert.
    _adjust_ious_for_crowd(ious, preds_area, gts_crowd_sorted, gts_area_sorted)

    ignore_preds = np.zeros_like(ignore_preds_area)

    for i in range(matched.shape[0]):
        p_ind = sort_preds[i]
        # Try mathing with normal gts
        if n_gts > 0:
            best_match_ind = np.argmax(ious[p_ind, :n_gts])
            if ious[p_ind, best_match_ind] > iou_threshold:
                ious[:, best_match_ind] = -1
                matched[p_ind] = sort_gt[best_match_ind]
                continue

        # Try matching with ignored gts
        if n_gts < len(gts_area):
            ignore_match_ind = np.argmax(ious[p_ind, n_gts:]) + n_gts
            if ious[p_ind, ignore_match_ind] > iou_threshold:
                # Crowd gts can be matched by multiple detections
                if not gts_crowd_sorted[ignore_match_ind]:
                    ious[:, ignore_match_ind] = -1
                matched[p_ind] = sort_gt[ignore_match_ind]
                ignore_preds[p_ind] = True

    ignore_preds = ignore_preds | ((matched == -1) & (ignore_preds_area))

    return matched, ignore_preds, gts_ignore, n_gts


@numba.njit(  # type: ignore[misc]
    types.Tuple((types.int32[::1], types.bool_[::1], types.bool_[::1]))(
        types.float32[:, ::1],
        types.float32[:, ::1],
        types.bool_[::1],
        types.DictType(types.int64, types.float32[:, ::1]),
        types.float32[::1],
        types.int32[:, ::1],
        types.Tuple((types.float32, types.float32)),
        types.float32,
    ),
    parallel=False,
)
def _evaluate_dataset(
    preds_bbox: npt.NDArray[np.float32],
    gts_bbox: npt.NDArray[np.float32],
    gts_crowd: npt.NDArray[np.bool_],
    ious: dict[int, npt.NDArray[np.float32]],
    preds_conf: npt.NDArray[np.float32],
    img_ind_corr: npt.NDArray[np.float32],
    area_range: tuple[float, float],
    iou_threshold: float,
) -> tuple[npt.NDArray[np.int32], npt.NDArray[np.bool_], npt.NDArray[np.bool_]]:
    """
    Calculate the precision-recall curve.

    Args:
        preds_bbox: A `[N, 4]` array of prediction bounding boxes in xywh format
        gts_bbox: A `[M, 4]` array of ground truth bounding boxes in xywh format
        gts_crowd: A `[M,]` array determining if a ground truth is a crowd or not.
        ious: A dict with keys being image indices, and values beind 2D numpy arrays,
          containing IoU similarity scores between predictions and ground truths for
          each image.
        preds_conf: A `[N,]` array of prediction confidence scores.
        img_ind_corr: An `[K, 4]` array, that for each image (in the union of images
            from both hypotheses and ground truths) contains the start and end indices
            for that image in both hypotheses and ground truths. A row is of the form
            `hyp_start, hyp_end, gt_start, gt_end`. If `hyp_end` or `gt_end` is 0, that
            means that this image does not appear in hypotheses or ground truths,
            respectivelly.
        area_range: A tuple `(lower_limit, upper_limit)` of limits for bounding box
            areas. Ground truths with area outside of this range will be ignored, as
            will the predictions matched to them and unmatched predictions with area
            outside of this range
        iou_threshold: The lower threshold for an IOU between a prediction and a
            bounding box that is required for establishing a match

    Returns:
        A tuple with 3 elements:
            - an integer array of shape `[N, ]`, indicating the index of the ground
              truth that the prediction was matched to. Unmatched predictions will have
              the matched index of `-1`.
            - a boolean array of shape `[N, ]`, indicating whether the prediction was
              ignored or not
            - a boolean array of shape `[M, ]`, indicating whether the ground truth was
              ignored or not

    """
    det_ignored = np.zeros_like(preds_conf, dtype=np.bool_)
    gts_ignored = np.zeros((gts_bbox.shape[0],), dtype=np.bool_)
    det_matched = np.full_like(preds_conf, -1, dtype=np.int32)

    for img in numba.prange(len(img_ind_corr)):
        preds_start, preds_end = img_ind_corr[img, :2]
        gts_start, gts_end = img_ind_corr[img, 2:]

        if preds_end == 0 and gts_end == 0:
            continue

        img_ious = np.zeros((0, 0), dtype=np.float32)
        if img in ious:
            img_ious = ious[img]

        (matched_img, ignored_dets_img, ignored_gts_img, _) = _evaluate_image(
            preds_bbox[preds_start:preds_end],
            gts_bbox[gts_start:gts_end],
            gts_crowd[gts_start:gts_end],
            img_ious,
            preds_conf=preds_conf[preds_start:preds_end],
            area_range=area_range,
            iou_threshold=iou_threshold,
        )

        if preds_end > 0:
            det_ignored[preds_start:preds_end] = ignored_dets_img

            # Add gts_start as image offset
            matched_img = matched_img + numba.int32(gts_start)
            matched_img[matched_img < gts_start] = -1
            det_matched[preds_start:preds_end] = matched_img

        if gts_end > 0:
            gts_ignored[gts_start:gts_end] = ignored_gts_img

    return det_matched, det_ignored, gts_ignored


def _check_compatibility(gt: Detections, hyp: Detections) -> None:
    if hyp.class_names != gt.class_names:
        raise ValueError(
            "`class_names` must be the same for ground truths and hypotheses"
        )

    if hyp.confs is None:
        raise ValueError("`confs` must be provided for hypotheses")


def _compute_ap(
    precision: npt.NDArray[np.float64] | None,
    recall: npt.NDArray[np.float64] | None,
    ap_interpolation: APInterpolation = APInterpolation.coco,
) -> float | None:
    if precision is None or recall is None:
        return None

    precision = np.concatenate([[0], precision])
    recall = np.concatenate([[0], recall])
    precision = np.maximum.accumulate(precision[::-1])[::-1]

    if ap_interpolation == "coco":
        thresholds = np.linspace(0.0, 1.00, 101, endpoint=True)
        r_inds = np.searchsorted(recall, thresholds, side="left")
        prec_a = np.concatenate([precision, [0]])
        return prec_a[r_inds].mean()  # type: ignore[no-any-return]

    if ap_interpolation == "pascal":
        return np.trapezoid(precision, recall)  # type: ignore[no-any-return]

    # No matching ap_interpolation
    raise ValueError(f"Unknown interpolation {ap_interpolation}")


def compute_metrics(
    gt: Detections,
    hyp: Detections,
    iou_threshold: float,
    area_range: tuple[float, float] = (0.0, float("inf")),
    ap_interpolation: APInterpolation = APInterpolation.coco,
) -> COCOResults:
    """
    Compute COCO metrics (mAP, precision and recall).

    The metrics are based on the [detection metrics](https://cocodataset.org/#detection-eval)
    of the COCO project. The metrics implemented here replicate the matching mechanism
    of the official ones, with the main difference being that the `"crowd"` attribute
    of ground truth objects is not taken into account here.

    Args:
        gt: Ground truth detections.
        hyp: Hypotheses detections.
        iou_threshold: IoU threshold for matching.
        area_range: The upper and lower threshold for object area. Objects outside
            of this range will be ignored.
        ap_interpolation: The method to use for interpolating the average precision
            curve. The `coco` method uses 101 equally spaced point, while the
            `pascal` method directly computes the area under the (max) curve. In
            practice the difference between the two methods will be minimal.

    Returns:
        A dictionary that contains the mean average precision metrics for the whole
        dataset (averaged class APs), as well as the results for each class (AP,
        precision and recall).

    """
    _check_compatibility(gt, hyp)
    class_results: dict[str, COCOResult] = {}

    for i, cls_name in enumerate(gt.class_names):
        hyp_cls = hyp.filter(hyp.classes == i)
        gt_mask = gt.classes == i
        gt_cls = gt.filter(gt_mask)
        gt_crowd_cls = gt_cls.iscrowd
        if gt_crowd_cls is None:
            gt_crowd_cls = np.zeros((len(gt_cls.bboxes),), dtype=np.bool_)

        img_ind_corr = utils.match_images(gt_cls, hyp_cls)
        ious = utils.compute_ious(hyp_cls.bboxes, gt_cls.bboxes, img_ind_corr)

        hyp_matched, hyp_ignored, gts_ignored = _evaluate_dataset(
            preds_bbox=hyp_cls.bboxes,
            gts_bbox=gt_cls.bboxes,
            gts_crowd=gt_crowd_cls,
            ious=ious,
            preds_conf=hyp_cls.confs,
            img_ind_corr=img_ind_corr,
            area_range=area_range,
            iou_threshold=iou_threshold,
        )
        n_gts = len(gt_cls.bboxes) - gts_ignored.sum()

        if hyp_cls.confs is None:
            msg = "hyp_cls.confs should not be None"
            raise ValueError(msg)

        precision, recall = prec_recall_curve(
            hyp_matched[~hyp_ignored] != -1, hyp_cls.confs[~hyp_ignored], n_gts
        )

        if precision is None or recall is None:
            class_results[cls_name] = {
                "ap": None,
                "recall": None,
                "precision": None,
                "n_gts": n_gts,
            }
        else:
            ap = _compute_ap(precision, recall, ap_interpolation)

            class_results[cls_name] = {
                "ap": ap,
                "recall": recall[-1],
                "precision": precision[-1],
                "n_gts": n_gts,
            }

    # Aggregate metrics
    aps = np.array([res["ap"] for res in class_results.values()], dtype=float)
    mean_ap = _nonemean(aps)
    metrics: COCOResults = {"mean_ap": mean_ap, "class_results": class_results}
    return metrics


_IOU_50_INDEX = 0
_IOU_75_INDEX = 5


def compute_coco_summary(
    gt: Detections,
    hyp: Detections,
    sizes: Mapping[str, tuple[float, float]] = {
        "small": (0.0, 32**2),
        "medium": (32**2, 96**2),
        "large": (96**2, float("inf")),
    },
    ap_interpolation: APInterpolation = APInterpolation.coco,
) -> COCOSummaryResults:
    """
    Compute COCO summary metrics.

    This computes the standard COCO summary metrics:
    * mAP (overall, and for each class)
    * mAP at IoU threshold of 0.5 and 0.75 (overall and for each class)
    * mAP for different sizes of objects (overall and for each class)

    The metrics are based on the [detection metrics](https://cocodataset.org/#detection-eval)
    of the COCO project. The metrics implemented here replicate the matching mechanism
    of the official ones, with the main difference being that the `"crowd"` attribute
    of ground truth objects is not taken into account here.

    Computing the summary metrics via this method is more efficient than using
    `compute_metrics` to compute them individually, as some intermediate steps
    (matching images, computing IoUs) can be shared between metrics.

    Args:
        gt: Ground truth detections
        hyp: Hypotheses detections
        sizes: A dictionary with size names as keys and their area ranges as values.
        ap_interpolation: The method to use for interpolating the average precision
            curve. The `coco` method uses 101 equally spaced point, while the
            `pascal` method directly computes the area under the (max) curve. In
            practice the difference between the two methods will be minimal.

    Returns:
        A dictionary with all overall and per class summary metrics.

    """
    _check_compatibility(gt, hyp)

    sizes_all = {"_all": (0.0, float("inf"))} | sizes  # type: ignore[operator]
    ap_results: dict[tuple[str, str, int], float | None] = {}

    for i_cls, cls_name in enumerate(gt.class_names):
        hyp_cls = hyp.filter(hyp.classes == i_cls)
        gt_cls = gt.filter(gt.classes == i_cls)

        gt_crowd_cls = gt_cls.iscrowd
        if gt_crowd_cls is None:
            gt_crowd_cls = np.zeros((len(gt_cls.bboxes),), dtype=np.bool_)

        img_ind_corr = utils.match_images(gt_cls, hyp_cls)

        ious = utils.compute_ious(hyp_cls.bboxes, gt_cls.bboxes, img_ind_corr)

        for i_thr, iou_threshold in enumerate(_BASE_50_95_THRESHOLDS):
            for size_name, area_range in sizes_all.items():
                hyp_matched, hyp_ignored, gts_ignored = _evaluate_dataset(
                    preds_bbox=hyp_cls.bboxes,
                    gts_bbox=gt_cls.bboxes,
                    gts_crowd=gt_crowd_cls,
                    ious=ious,
                    preds_conf=hyp_cls.confs,
                    img_ind_corr=img_ind_corr,
                    area_range=area_range,
                    iou_threshold=iou_threshold,
                )
                n_gts = len(gt_cls.bboxes) - gts_ignored.sum()

                if hyp_cls.confs is None:
                    msg = "hyp_cls.confs should not be None"
                    raise ValueError(msg)

                precision, recall = prec_recall_curve(
                    hyp_matched[~hyp_ignored] != -1,
                    hyp_cls.confs[~hyp_ignored],
                    n_gts,
                )

                ap_results[(cls_name, size_name, i_thr)] = _compute_ap(
                    precision, recall, ap_interpolation
                )

    # Aggregate metricss
    aps = np.array(list(ap_results.values()), dtype=np.float64)
    cls_arr = np.array([k[0] for k in ap_results])
    size_arr = np.array([k[1] for k in ap_results])
    iou_t_arr = np.array([k[2] for k in ap_results])

    results: dict[str, Any] = {}
    results["mean_ap"] = _nonemean(aps[size_arr == "_all"])
    results["ap_50"] = _nonemean(
        aps[(iou_t_arr == _IOU_50_INDEX) & (size_arr == "_all")]
    )
    results["ap_75"] = _nonemean(
        aps[(iou_t_arr == _IOU_75_INDEX) & (size_arr == "_all")]
    )

    results["mean_ap_per_class"] = {}
    results["ap_50_per_class"] = {}
    results["ap_75_per_class"] = {}

    for cls in gt.class_names:
        results["mean_ap_per_class"][cls] = _nonemean(
            aps[(size_arr == "_all") & (cls_arr == cls)]
        )
        results["ap_50_per_class"][cls] = _nonemean(
            aps[(size_arr == "_all") & (cls_arr == cls) & (iou_t_arr == _IOU_50_INDEX)]
        )
        results["ap_75_per_class"][cls] = _nonemean(
            aps[(size_arr == "_all") & (cls_arr == cls) & (iou_t_arr == _IOU_75_INDEX)]
        )

    results["mean_ap_sizes"] = {}
    for size_name in sizes:
        results["mean_ap_sizes"][size_name] = _nonemean(aps[size_arr == size_name])

    results["mean_ap_sizes_per_class"] = {}
    for cls in gt.class_names:
        results["mean_ap_sizes_per_class"][cls] = {}
        for size_name in sizes:
            results["mean_ap_sizes_per_class"][cls][size_name] = _nonemean(
                aps[(size_arr == size_name) & (cls_arr == cls)]
            )

    return cast("COCOSummaryResults", results)


def confusion_matrix(
    gt: Detections,
    hyp: Detections,
    iou_threshold: float,
    area_range: tuple[float, float] = (0.0, float("inf")),
) -> npt.NDArray[np.int32]:
    """
    Compute the confusion matrix.

    This method computes the confusion matrix, which matches objects in hypotheses
    and ground truth in a class-agnostic way, and then computes the number of
    objects in hypotheses of class X that were matched with object of class Y in
    ground truths. Number of unmatched objects in hypotheses and ground truths is
    also calculated.

    Args:
        gt: Ground truth detections. It must contain `image_names` property, which
            is used to match the images between `gt` and `hyp`. If it contains
            `class_names`, they must match the ones from `hyp`.
        hyp: Hypotheses (prediction) detections. It must contain `image_names`
            property, which is used to match the images between `gt` and `hyp`. If
            it contains `class_names`, they must match the ones from `gt`.
        iou_threshold: IoU threshold for matching hypotheses objects to ground
            truth.
        area_range: A tuple of `(lower_limit, upper_limit)` floats that set
            the lower and upper threshold for bound box area - detections with area
            outside of this range will be ignored.

    Returns:
        A `[C + 1, C + 1]` matrix, where `C` is the number of classes, and the entry
        at `[i, j]` denotes the number of hypotheses with class index `i` that were
        matched to a ground truth with the class index `j`. If the row or column
        index is `C` (last one), then this corresponds to the number of hypotheses
        or ground truths, respectively, that were not matched.

    """
    _check_compatibility(gt, hyp)

    img_ind_corr = utils.match_images(gt, hyp)
    ious = utils.compute_ious(hyp.bboxes, gt.bboxes, img_ind_corr)

    gt_crowd = gt.iscrowd
    if gt_crowd is None:
        gt_crowd = np.zeros((len(gt.bboxes),), dtype=np.bool_)

    det_matched, det_ignored, gts_ignored = _evaluate_dataset(
        preds_bbox=hyp.bboxes,
        gts_bbox=gt.bboxes,
        gts_crowd=gt_crowd,
        ious=ious,
        preds_conf=hyp.confs,
        img_ind_corr=img_ind_corr,
        area_range=area_range,
        iou_threshold=iou_threshold,
    )

    det_matched = det_matched[~det_ignored]
    det_classes = hyp.classes[~det_ignored]

    n_classes = len(gt.class_names)
    return utils.confusion_matrix(  # type: ignore[no-any-return]
        det_matched, gts_ignored, det_classes, gt.classes, n_classes
    )



"""Multi Object Tracking (MOT) metrics."""

__all__ = [
    "CLEARMOTResults",
    "HOTAResults",
    "IDResults",
    "MOTMetrics",
    "MOTMetricsResults",
    "calculate_clearmot_metrics",
    "calculate_hota_metrics",
    "calculate_id_metrics",
]
from .clearmot import CLEARMOTResults, calculate_clearmot_metrics
from .hota import HOTAResults, calculate_hota_metrics
from .identity import IDResults, calculate_id_metrics
from .motmetrics import MOTMetrics, MOTMetricsResults



"""CLEARMOT MOT metrics."""

import logging
import typing as t

import numpy as np
import numpy.typing as npt
from scipy.optimize import linear_sum_assignment

from evaldet.tracks import FrameTracks, Tracks

logger = logging.getLogger(__name__)


class CLEARMOTResults(t.TypedDict):
    """
    A typed dictionary for storing the results of the CLEARMOT metric evaluation.
    """

    MOTP: float
    MOTA: float
    FP_CLEAR: float
    FN_CLEAR: float
    IDSW: float


def _update_frame_matches(
    matching: dict[int, int],
    matching_persist: dict[int, int],
    gt: FrameTracks,
    hyp: FrameTracks,
    iou_frame: npt.NDArray[np.float32],
    dist_threshold: float,
    matches_dist: list[float],
    mismatches: int,
) -> int:
    """
    Update `matching` for the current frame based on IOU distances and
    return updated `mismatches`.
    """
    # Remove matches that do not exist in this frame's GT or Hyp
    for key in list(matching.keys()):
        if matching[key] not in hyp.ids or key not in gt.ids:
            del matching[key]

    # Build index dictionaries for GT and hypotheses
    id_ind_dict_gt = {_id: ind for ind, _id in enumerate(gt.ids)}
    id_ind_dict_hyp = {_id: ind for ind, _id in enumerate(hyp.ids)}

    # Remove matches above the distance threshold; store valid matches' distances
    for gt_id, hyp_id in tuple(matching.items()):
        dist_match = iou_frame[id_ind_dict_gt[gt_id], id_ind_dict_hyp[hyp_id]]
        if dist_match > dist_threshold:
            del matching[gt_id]
        else:
            matches_dist.append(dist_match)

    # Identify which GT and Hyp IDs remain unmatched
    match_gt_ids = tuple(matching.keys())
    match_hyp_ids = tuple(matching.values())
    inds_nm_gt = np.where(np.isin(gt.ids, match_gt_ids, invert=True))[0]
    inds_nm_hyp = np.where(np.isin(hyp.ids, match_hyp_ids, invert=True))[0]
    dist_matrix = iou_frame[np.ix_(inds_nm_gt, inds_nm_hyp)]

    # Assign remaining unmatched IDs using LAP
    for row_ind, col_ind in zip(*linear_sum_assignment(dist_matrix), strict=True):
        if dist_matrix[row_ind, col_ind] < dist_threshold:
            matching[gt.ids[inds_nm_gt[row_ind]]] = hyp.ids[inds_nm_hyp[col_ind]]
            matches_dist.append(dist_matrix[row_ind, col_ind])

    # Check if mismatches (ID switches) have occurred
    for key, matched_track in matching.items():
        if key in matching_persist and matched_track != matching_persist[key]:
            mismatches += 1

    return mismatches


def calculate_clearmot_metrics(
    ground_truth: Tracks,
    hypotheses: Tracks,
    ious: dict[int, npt.NDArray[np.float32]],
    dist_threshold: float = 0.5,
) -> CLEARMOTResults:
    """
    Calculate CLEARMOT metrics.

    Args:
        ground_truth: Ground truth tracks.
        hypotheses: Hypotheses tracks.
        ious: A dictionary where keys are frame numbers (indices), and values
            are numpy matrices of IOU distances between detection in ground truth and
            hypotheses for that frame. IOUs must be present for all frames that are
            present in both ground truth and hypotheses.
        dist_threshold: The distance threshold for the computation of metrics, used to
            determine whether a matching between two tracks persist, and whether to
            start a matching based on distance between two detections.

    Returns:
        A dictionary containing CLEARMOT metrics:

            - MOTA (MOT Accuracy)
            - MOTP (MOT Precision)
            - FP (false positives)
            - FN (false negatives)
            - IDS (identity switches/mismatches)

    """
    all_frames = sorted(ground_truth.frames.union(hypotheses.frames))

    false_negatives = 0
    false_positives = 0
    mismatches = 0
    ground_truths = 0

    matches_dist: list[float] = []
    matching: dict[int, int] = {}

    # This is the persistent matching dictionary, used to check for mismatches
    # when a previously matched hypothesis is re-matched with a ground truth
    matching_persist: dict[int, int] = {}

    for frame in all_frames:
        if frame not in ground_truth:
            matching = {}
            false_positives += len(hypotheses[frame].ids)
        elif frame not in hypotheses:
            matching = {}
            false_negatives += len(ground_truth[frame].ids)
            ground_truths += len(ground_truth[frame].ids)
        else:
            hyp, gt = hypotheses[frame], ground_truth[frame]

            mismatches = _update_frame_matches(
                matching=matching,
                matching_persist=matching_persist,
                gt=gt,
                hyp=hyp,
                iou_frame=ious[frame],
                dist_threshold=dist_threshold,
                matches_dist=matches_dist,
                mismatches=mismatches,
            )

            # Compute false positives and false negatives
            false_negatives += len(gt.ids) - len(matching)
            false_positives += len(hyp.ids) - len(matching)

            ground_truths += len(gt.ids)
            matching_persist.update(matching)

    if not matches_dist:
        logger.warning("No matches were made, MOPT will be np.nan")
        motp = np.nan
    else:
        motp = sum(matches_dist) / len(matches_dist)

    mota = 1 - (false_negatives + false_positives + mismatches) / ground_truths

    return {
        "MOTP": motp,
        "MOTA": mota,
        "FP_CLEAR": false_positives,
        "FN_CLEAR": false_negatives,
        "IDSW": mismatches,
    }



"""Various utilitis for computing object detection metrics."""

import numba as nb
import numpy as np
import numpy.typing as npt
from numba import types

from evaldet import Detections
from evaldet.dist import iou_dist


@nb.njit(  # type: ignore[misc]
    types.DictType(types.int64, types.float32[:, ::1])(
        types.float32[:, ::1],
        types.float32[:, ::1],
        types.int32[:, ::1],
    )
)
def compute_ious(
    hyp_bboxes: npt.NDArray[np.float32],
    gt_bboxes: npt.NDArray[np.float32],
    image_ind_corr: npt.NDArray[np.int32],
) -> types.DictType:
    """
    Compute the IoUs for each image.

    This computes the IoUs between bounding boxes in hypotheses and ground truth for
    each image present in both

    Args:
        hyp_bboxes: An `[N,4]` array of hypotheses bounding boxes, where each row is of
            the format `xywh`.
        gt_bboxes: An `[M,4]` array of ground truth bounding boxes, where each row is of
            the format `xywh`.
        image_ind_corr: An `[K,4]` array where each row is of the form
            `[hyp_start, hyp_end, gt_start, gt_end]`. The start/end indices indicate at
            which index do elements of the frame start or end in the bounding box array,
            for hypotheses and indices. If the end index is 0, this means that this
            frame is not present in hypotheses/ground truths

    Returns:
        A numba dictionary where key is the image index, corresponding to the row index
        in `image_ind_corr`, and the key is a `[H, G]` array of IoUs, where rows
        correspond to hypotheses and coulmns to ground truths on that image.

    """
    ious = {}
    common_images = np.nonzero(image_ind_corr[:, 3] * image_ind_corr[:, 1])[0]

    for img_id in common_images:
        hyp_start, hyp_end = image_ind_corr[img_id, :2]
        hyp_img_bboxes = hyp_bboxes[hyp_start:hyp_end]

        gt_start, gt_end = image_ind_corr[img_id, 2:]
        gt_img_bboxes = gt_bboxes[gt_start:gt_end]

        ious[img_id] = 1 - iou_dist(hyp_img_bboxes, gt_img_bboxes)

    return ious


def match_images(gt: Detections, hyp: Detections) -> npt.NDArray[np.int32]:
    """
    Match images between ground truths and hypotheses.

    This creates an array with the start and end indices of the same image for
    detections and hypotheses.

    Args:
        gt: Ground truth detections
        hyp: Hypotheses (predictions) detections

    Returns:
        An array of shape `[N, 4]`, where each row is of the form
        `[hyp_start, hyp_end, gt_start, gt_end]`. The start/end indices indicate at
        which index do elements of the frame start. or end (for example for bboxes),
        for hypotheses and indices. If the end index is 0, this means that this frame is
        not present in hypotheses/ground truths

    """
    gt_img_dict = {name: i for i, name in enumerate(gt.image_names)}
    hyp_img_dict = {name: i for i, name in enumerate(hyp.image_names)}

    all_images = sorted(set(gt.image_names).union(hyp.image_names))
    image_ind_corr = np.zeros((len(all_images), 4), dtype=np.int32)

    for i, img in enumerate(all_images):
        if img in gt_img_dict:
            img_id = gt_img_dict[img]
            if img_id in gt.image_ind_dict:
                image_ind_corr[i, 2:] = gt.image_ind_dict[img_id]

        if img in hyp_img_dict:
            img_id = hyp_img_dict[img]
            if img_id in hyp.image_ind_dict:
                image_ind_corr[i, :2] = hyp.image_ind_dict[img_id]

    return image_ind_corr


@nb.njit(  # type: ignore[misc]
    nb.int32[:, ::1](
        nb.int32[::1],
        nb.boolean[::1],
        nb.int32[::1],
        nb.int32[::1],
        nb.int32,
    ),
)
def confusion_matrix(
    matching: npt.NDArray[np.int32],
    gt_ignored: npt.NDArray[np.bool_],
    hyp_classes: npt.NDArray[np.int32],
    gt_classes: npt.NDArray[np.int32],
    n_classes: int,
) -> npt.NDArray[np.int32]:
    """
    Compute the confusion matrix.

    This method computes the confusion matrix, showing the number of objects in
    hypotheses of class X that were matched with object of class Y in ground truths.

    Args:
        matching: A `[N,]` shaped integer array, where entry at `i`-th position denotes
            the index of the ground truth that the `i`-th hypothesis is matched with.
        gt_ignored: A `[M,]` array, denoting whether a ground truth is ignored or not.
        hyp_classes: A `[N,]` array, denoting the class index of each hypothesis.
        gt_classes: A `[M,]` array, denoting the class index of each ground truth.
        n_classes: Total number of classes.

    Returns:
        A `[n_classes + 1, n_classes + 1]` matrix, where the entry  at `[i, j]` denotes
        the number of hypotheses with class index `i` that were matched to a ground
        truth with the class index `j`. If the row or column index is `n_classes`
        (last one), then this corresponds to the number of hypotheses or ground truths,
        respectively, that were not matched.

    """
    conf_matrix = np.zeros((n_classes + 1, n_classes + 1), dtype=np.int32)

    # First fill in totals for gts
    gt_non_ignored_classes = gt_classes[~gt_ignored]
    for i in range(gt_non_ignored_classes.shape[0]):
        conf_matrix[-1, gt_non_ignored_classes[i]] += 1

    for i in range(matching.shape[0]):
        r = hyp_classes[i]
        gt_ind = matching[i]

        if gt_ind > -1 and gt_ignored[gt_ind] is True:
            c = -1
        elif gt_ind > -1:
            c = gt_classes[gt_ind]

            # Subtract match from gt totals
            conf_matrix[-1, c] -= 1
        else:
            c = -1

        conf_matrix[r, c] += 1

    return conf_matrix



"""Module with objects for computing MOT metrics from various metric families."""

import logging
import typing as t

import numpy as np
import numpy.typing as npt

from evaldet.dist import iou_dist
from evaldet.tracks import Tracks
from evaldet.utils import timer

from .clearmot import CLEARMOTResults, calculate_clearmot_metrics
from .hota import HOTAResults, calculate_hota_metrics
from .identity import IDResults, calculate_id_metrics


class MOTMetricsResults(t.TypedDict):
    """
    A typed dictionary for storing the results of various MOT metrics evaluations.
    """

    clearmot: CLEARMOTResults | None
    id: IDResults | None
    hota: HOTAResults | None


class MOTMetrics:
    """
    The class for computing MOT metrics.

    To compute the metrics, use the ``compute`` method of this class, it will compute
    all the required MOT metrics.

    The reason for a single entrypoint for MOT computation is so that metrics can
    efficiently share pre-computed IoU distances.
    """

    _ious_dict: dict[int, npt.NDArray[np.float32]]

    def __init__(
        self, clearmot_dist_threshold: float = 0.5, id_dist_threshold: float = 0.5
    ) -> None:
        """
        Initialize the object.

        Args:
            clearmot_dist_threshold: The distance threshold for the computation of
                CLEARMOT metrics, used to determine whether a matching between two
                tracks persist, and whether to start a matching based on distance
                between two detections.
            id_dist_threshold: The distance threshold for the computation of the
                ID metrics - used to determine whether to match two objects.

        """
        self._clearmot_dist_threshold = clearmot_dist_threshold
        self._id_dist_threshold = id_dist_threshold

        self._logger = logging.getLogger(self.__class__.__name__)

    def compute(
        self,
        ground_truth: Tracks,
        hypotheses: Tracks,
        clearmot_metrics: bool = False,
        id_metrics: bool = False,
        hota_metrics: bool = True,
    ) -> MOTMetricsResults:
        """
        Compute multi-object tracking (MOT) metrics.

        Right now, the following metrics can be computed

            - CLEARMOT metrics

                - MOTA (MOT Accuracy)
                - MOTP (MOT Precision)
                - FP (false positives)
                - FN (false negatives)
                - IDS (identity switches/mismatches)

            - ID metrics

                - IDP (ID Precision)
                - IDR (ID Recall)
                - IDF1 (ID F1)
                - IDFP (ID false positives)
                - IDFN (ID false negatives)
                - IDTP (ID true positives)

            - HOTA metrics. Note that I use the matching algorithm from the paper,
            which differs from what the official repository (TrackEval) is using - see
            [this issue](https://github.com/JonathonLuiten/TrackEval/issues/22)
            for more details

                - HOTA (average and per-alpha)
                - AssA (average and per-alpha)
                - DetA (average and per-alpha)
                - LocA
                - DetPr (per-alpha)
                - DetRec (per-alpha)
                - DetTP (per-alpha)
                - DetFP (per-alpha)
                - DetFN (per-alpha)
                - AssPr (per-alpha)
                - AssRec (per-alpha)

        Args:
            ground_truth: Ground truth tracks.
            hypotheses: Hypotheses tracks.
            clearmot_metrics: Whether to compute the CLEARMOT metrics.
            id_metrics: Whether to compute the ID metrics.
            hota_metrics: Whether to compute the HOTA metrics.

        Returns:
            A dictionary of computed metrics. Metrics are saved under the key of their
            metric family (`"clearmot"`, `"id"`, `"hota"`).

        """
        if not (clearmot_metrics or id_metrics or hota_metrics):
            msg = "You must select some metrics to compute."
            raise ValueError(msg)

        if not len(ground_truth):
            msg = "No objects in ``ground_truths``, nothing to compute."
            raise ValueError(msg)

        with timer.timer(self._logger, "Precompute IoU"):
            self._ious_dict = _compute_ious(ground_truth, hypotheses)

        if clearmot_metrics:
            with timer.timer(self._logger, "Compute CLEARMOT Metrics"):
                clrmt_metrics = calculate_clearmot_metrics(
                    ground_truth,
                    hypotheses,
                    self._ious_dict,
                    self._clearmot_dist_threshold,
                )
        else:
            clrmt_metrics = None

        if id_metrics:
            with timer.timer(self._logger, "Compute ID Metrics"):
                id_metrics_res = calculate_id_metrics(
                    ground_truth, hypotheses, self._ious_dict, self._id_dist_threshold
                )
        else:
            id_metrics_res = None

        if hota_metrics:
            with timer.timer(self._logger, "Compute HOTA Metrics"):
                hota_metrics_res = calculate_hota_metrics(
                    ground_truth,
                    hypotheses,
                    self._ious_dict,
                )
        else:
            hota_metrics_res = None

        # Remove IoU matrix to release memory
        del self._ious_dict

        return MOTMetricsResults(
            clearmot=clrmt_metrics, id=id_metrics_res, hota=hota_metrics_res
        )


def _compute_ious(
    tracks_1: Tracks, tracks_2: Tracks
) -> dict[int, npt.NDArray[np.float32]]:
    all_frames = sorted(set(tracks_1.frames).intersection(tracks_2.frames))

    ious: dict[int, npt.NDArray[np.float32]] = {}
    for frame in all_frames:
        ious_f = iou_dist(tracks_1[frame].bboxes, tracks_2[frame].bboxes)
        ious[frame] = ious_f

    return ious



"""HOTA family of MOT metrics."""

import collections as co
import typing as t

import numpy as np
import numpy.typing as npt
from scipy.optimize import linear_sum_assignment

from evaldet.tracks import Tracks

from .utils import create_coo_array

_EPS = 1 / 1000


class HOTAResults(t.TypedDict):
    """
    A typed dictionary for storing the results of the HOTA metric evaluation.
    """

    HOTA: float
    DetA: float
    AssA: float
    LocA: float
    alphas_HOTA: npt.NDArray[np.float32]
    HOTA_alpha: npt.NDArray[np.float32]
    DetA_alpha: npt.NDArray[np.float32]
    AssA_alpha: npt.NDArray[np.float32]
    LocA_alpha: npt.NDArray[np.float32]

    DetPr_alpha: npt.NDArray[np.float32]
    DetRec_alpha: npt.NDArray[np.float32]
    DetTP_alpha: npt.NDArray[np.int64]
    DetFN_alpha: npt.NDArray[np.int64]
    DetFP_alpha: npt.NDArray[np.int64]

    AssPr_alpha: npt.NDArray[np.float32]
    AssRec_alpha: npt.NDArray[np.float32]


def calculate_hota_metrics(  # noqa: PLR0915
    ground_truth: Tracks, hypotheses: Tracks, ious: dict[int, npt.NDArray[np.float32]]
) -> HOTAResults:
    """
    Calculate HOTA metrics.

    Args:
        ground_truth: Ground truth tracks.
        hypotheses: Hypotheses tracks.
        ious: A dictionary where keys are frame numbers (indices), and values
            are numpy matrices of IOU distances between detection in ground truth and
            hypotheses for that frame. IOUs must be present for all frames that are
            present in both ground truth and hypotheses.

    Returns:
        A dictionary containing HOTA metrics. Note that I use the matching algorithm
        from the paper, which differs from what the official repository (TrackEval) is
        using - see [this issue](https://github.com/JonathonLuiten/TrackEval/issues/22)
        for more details. The metrics returned are:

            - HOTA (average and per-alpha)
            - AssA (average and per-alpha)
            - DetA (average and per-alpha)
            - LocA
            - DetPr (per-alpha)
            - DetRec (per-alpha)
            - DetTP (per-alpha)
            - DetFP (per-alpha)
            - DetFN (per-alpha)
            - AssPr (per-alpha)
            - AssRec (per-alpha)

    """
    alphas = np.arange(0.05, 0.96, 0.05)  # from 0.05 to 0.95 inclusive
    all_frames = sorted(set(ground_truth.frames).intersection(hypotheses.frames))

    gts = tuple(ground_truth.ids_count.keys())
    gts_counts = tuple(ground_truth.ids_count.values())
    gts_id_ind_dict = {_id: ind for ind, _id in enumerate(gts)}

    hyps = tuple(hypotheses.ids_count.keys())
    hyps_counts = tuple(hypotheses.ids_count.values())
    hyps_id_ind_dict = {_id: ind for ind, _id in enumerate(hyps)}

    n_gt, n_hyp = len(gts), len(hyps)
    FP, FN = sum(hyps_counts), sum(gts_counts)

    DetAs = np.zeros_like(alphas)
    AssAs = np.zeros_like(alphas)
    LocAs = np.zeros_like(alphas)
    DetPrs = np.zeros_like(alphas)
    DetRecs = np.zeros_like(alphas)
    DetTPs = np.zeros_like(alphas)
    DetFPs = np.zeros_like(alphas)
    DetFNs = np.zeros_like(alphas)
    AssPrs = np.zeros_like(alphas)
    AssRecs = np.zeros_like(alphas)

    for a_ind, alpha in enumerate(alphas):
        # The arrays should all have the shape [n_gt, n_hyp]
        FPA_max = np.tile(hyps_counts, (n_gt, 1))
        FNA_max = np.tile(gts_counts, (n_hyp, 1)).T
        TPA_max_vals: dict[tuple[int, int], int] = co.defaultdict(int)

        FPA, FNA = FPA_max.copy(), FNA_max.copy()
        locs = 0.0  # Accumulator of similarities

        # Do the optimisitc matching - allow multiple matches per gt/hyp in the
        # same frame
        for frame in all_frames:
            dist_matrix = ious[frame]

            gt_frame_inds = [gts_id_ind_dict[_id] for _id in ground_truth[frame].ids]
            hyp_frame_inds = [hyps_id_ind_dict[_id] for _id in hypotheses[frame].ids]

            for row_ind, col_ind in np.argwhere(dist_matrix < alpha):
                TPA_max_vals[(gt_frame_inds[row_ind], hyp_frame_inds[col_ind])] += 1

        TPA_max = create_coo_array(TPA_max_vals, (n_gt, n_hyp)).toarray()

        # Compute optimistic A_max, to be used for actual matching
        A_max = TPA_max / (FNA_max + FPA_max - TPA_max)

        # Do the actual matching
        TPA_vals: dict[tuple[int, int], int] = co.defaultdict(int)
        for frame in all_frames:
            dist_matrix = ious[frame]
            dist_cost = (1 - dist_matrix) * _EPS

            gt_ids_f = ground_truth[frame].ids
            hyp_ids_f = hypotheses[frame].ids
            gt_frame_inds = [gts_id_ind_dict[_id] for _id in gt_ids_f]
            hyp_frame_inds = [hyps_id_ind_dict[_id] for _id in hyp_ids_f]

            opt_matrix = ((dist_matrix < alpha) / _EPS).astype(np.float64)
            opt_matrix += A_max[np.ix_(gt_frame_inds, hyp_frame_inds)]
            opt_matrix += dist_cost

            # Calculate matching as a LAP
            matching_inds = linear_sum_assignment(opt_matrix, maximize=True)
            for row_ind, col_ind in zip(*matching_inds, strict=True):
                if dist_matrix[row_ind, col_ind] < alpha:
                    TPA_vals[(gt_frame_inds[row_ind], hyp_frame_inds[col_ind])] += 1
                    locs += 1 - dist_matrix[row_ind, col_ind]

        TPA = create_coo_array(TPA_vals, (n_gt, n_hyp)).toarray()

        # Compute proper scores
        TP = TPA.sum()
        A = TPA / (FNA + FPA - TPA)
        DetAs[a_ind] = TP / (FN + FP - TP)
        AssAs[a_ind] = (TPA * A).sum() / max(TP, 1)

        DetTPs[a_ind] = TP
        DetFPs[a_ind] = FP - TP
        DetFNs[a_ind] = FN - TP

        DetPrs[a_ind] = TP / FP
        DetRecs[a_ind] = TP / FN

        AssPrs[a_ind] = (TPA * TPA / FPA).sum() / max(TP, 1)
        AssRecs[a_ind] = (TPA * TPA / FNA).sum() / max(TP, 1)

        # If no matches -> full similarity [strange default]
        LocAs[a_ind] = np.maximum(locs, 1e-10) / np.maximum(TP, 1e-10)

    HOTAs = np.sqrt(DetAs * AssAs)

    return {
        "HOTA": HOTAs.mean(),
        "DetA": DetAs.mean(),
        "AssA": AssAs.mean(),
        "LocA": LocAs.mean(),
        "alphas_HOTA": alphas,
        "HOTA_alpha": HOTAs,
        "DetA_alpha": DetAs,
        "AssA_alpha": AssAs,
        "LocA_alpha": LocAs,
        "AssPr_alpha": AssPrs,
        "AssRec_alpha": AssRecs,
        "DetPr_alpha": DetPrs,
        "DetRec_alpha": DetRecs,
        "DetFN_alpha": DetFNs,
        "DetFP_alpha": DetFPs,
        "DetTP_alpha": DetTPs,
    }
